import os
import pathlib
from typing import Any, Generator, List, Optional, Tuple
from unittest import mock

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from click.testing import CliRunner

from gcs_migration.commands.gcs_migration import main
from gcs_migration.models import MigrationPlan, MyProxyConfiguration


@pytest.fixture
def mock_map_identities_environment(
    mock_get_transfer_api: Any,
    mock_get_auth_api: Any,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    fs: FakeFilesystem,
) -> Generator[MigrationPlan, None, None]:
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        mock_migration_load.return_value = migration_plan_cmd_create
        migration_plan_cmd_create.root = pathlib.Path(".")
        os.mkdir("migration_plan")
        yield migration_plan_cmd_create


@pytest.mark.parametrize(
    "use_host_cert_subject,domains,local_user,expect",
    [
        (True, ["globus.org"], "user", 0),
        (True, ["globus.org", "example.org"], "user", 0),
        (True, ["globus.org"], "unknownuser", 0),
        (False, ["globus.org"], "user", 1),
    ],
)
def test_identity_mapping_myproxy(
    mock_map_identities_environment: MigrationPlan,
    test_host_cert_subject: str,
    test_user_cert_subject: str,
    use_host_cert_subject: bool,
    domains: List[str],
    local_user: str,
    expect: int,
) -> None:
    migration_plan = mock_map_identities_environment

    if use_host_cert_subject:
        myproxy_ca_subject = test_host_cert_subject
    else:
        myproxy_ca_subject = test_user_cert_subject

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    assert migration_plan.guest_collections is not None
    assert migration_plan.guest_collections[0] is not None
    assert migration_plan.guest_collections[0].credential is not None
    # Set mapped user to be the a user with an auth identity fixture
    migration_plan.guest_collections[0].credential.username = local_user

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "myproxy"]

    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.myproxy_callouts is not None
    assert migration_plan.gsi_authz_configuration.myproxy_callouts[0] is not None

    if expect == 0:
        assert migration_plan.guest_collections is not None
        assert len(migration_plan.guest_collections) > 0
        assert migration_plan.guest_collections[0].credential is not None
        migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 0

    args.extend(["-s", myproxy_ca_subject])
    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == expect

    if expect == 0:
        assert migration_plan.storage_gateway.allowed_domains is not None
        assert len(migration_plan.storage_gateway.allowed_domains) == len(domains)
        for d in domains:
            assert d in migration_plan.storage_gateway.allowed_domains
        for d in domains:
            assert (
                d in migration_plan.gsi_authz_configuration.myproxy_callouts[0].domains
            )
        if "unknown" in local_user:
            assert "Unable to resolve" in result.output
    else:
        assert not migration_plan.storage_gateway.allowed_domains


def test_identity_mapping_myproxy_inferred_subject(
    mock_map_identities_environment: MigrationPlan,
) -> None:
    migration_plan = mock_map_identities_environment
    domains = ["globus.org"]

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    assert migration_plan.guest_collections is not None
    assert migration_plan.guest_collections[0] is not None
    assert migration_plan.guest_collections[0].credential is not None
    # Set mapped user to be the a user with an auth identity fixture
    migration_plan.guest_collections[0].credential.username = "user"

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "myproxy"]

    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.myproxy_callouts is not None
    assert migration_plan.gsi_authz_configuration.myproxy_callouts[0] is not None

    assert migration_plan.guest_collections is not None
    assert len(migration_plan.guest_collections) > 0
    assert migration_plan.guest_collections[0].credential is not None
    migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 0

    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == 0


def test_identity_mapping_myproxy_inferred_subject_no_config(
    mock_map_identities_environment: MigrationPlan,
) -> None:
    migration_plan = mock_map_identities_environment
    domains = ["globus.org"]

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "myproxy"]

    assert migration_plan.gsi_authz_configuration is not None
    migration_plan.gsi_authz_configuration.callouts = []

    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == 1


def test_identity_mapping_myproxy_inferred_subject_multiple_configs(
    mock_map_identities_environment: MigrationPlan,
) -> None:
    migration_plan = mock_map_identities_environment
    domains = ["globus.org"]

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "myproxy"]

    new_subject = "/C=US/O=Globus/CN=Other Subject"

    extra_myproxy_callout = MyProxyConfiguration()
    extra_myproxy_callout.dn = new_subject

    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.callouts is not None
    migration_plan.gsi_authz_configuration.callouts.append(extra_myproxy_callout)

    for d in domains:
        args.extend(["-d", d])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == 1


@pytest.mark.parametrize(
    "domains",
    [["email.org"], ["unknown.org"]],
)
def test_unusable_domains(
    mock_map_identities_environment: MigrationPlan,
    domains: List[str],
) -> None:
    migration_plan = mock_map_identities_environment

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "myproxy"]

    for d in domains:
        args.extend(["-d", d])

    assert migration_plan.storage_gateway is not None
    assert migration_plan.storage_gateway.allowed_domains is None

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == 1

    assert migration_plan.storage_gateway is not None
    assert migration_plan.storage_gateway.allowed_domains is None
    assert migration_plan.storage_gateway.identity_mappings is None


def test_requires_logged_in(
    mock_not_logged_in: None,
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        main,
        ["identity-mapping", "myproxy", "--domain", "foo.bar"],
        catch_exceptions=False,
    )
    assert "you will have to login" in result.output
