from typing import Any, Callable, Dict, Iterator, List, Optional, cast
from unittest import mock
import uuid

from click.testing import CliRunner
import pytest
from tests.parameters import connector_parameter

from globus_sdk import IdentityMap
from gcs_migration.commands.s3_credential_create import s3_credential_create
from gcs_migration.models import MigrationPlan, S3Credential
from gcs_migration.parsers import GCSConfig, GridFTPOptionParser, S3MapEntry
from gcs_migration.typing.auth_documents import IdentityDocumentDict


@pytest.fixture(autouse=True)
def connector() -> str:
    return "s3"


@pytest.fixture(autouse=True)
def s3_process_user() -> Optional[str]:
    return None


@pytest.fixture(autouse=True)
def mock_idmap_single_batch(monkeypatch: Any) -> Iterator[None]:
    monkeypatch.setattr(IdentityMap, "_default_id_batch_size", 1)
    yield


@pytest.fixture
def mock_identity_document() -> Callable[[str], IdentityDocumentDict]:
    namespace = uuid.UUID("216eccf8-9c32-4051-8239-1a99bfffee9d")

    def _mock_identity_document(user: str, domain: str = None) -> IdentityDocumentDict:
        if not domain:
            domain = "globus.org"
        local_username = user
        globus_username = f"{user}@{domain}"
        return {
            "username": globus_username,
            "status": "used",
            "name": local_username,
            "id": str(uuid.uuid5(namespace, globus_username)),
            "identity_type": "login",
            "organization": "Globus",
            "email": globus_username,
            "identity_provider": "d58174ba-05bf-4f62-a482-94586742b9ab",
        }

    return _mock_identity_document


@pytest.mark.parametrize(
    "connector",
    [
        connector_parameter("blackpearl"),
        connector_parameter("ceph"),
        connector_parameter("file"),
        connector_parameter("hpss_local"),
    ],
)
def test_s3_credential_create_non_s3_storage_gateway(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    s3_map_entries: List[S3MapEntry],
    connector: str,
) -> None:
    """
    When called with a migration plan using the file dsi, the error should
    indicate that this is an s3-only command.
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump_credentials"):
        mock_migration_load.return_value = migration_plan_cmd_create
        runner = CliRunner()
        local_user = s3_map_entries[1].username
        result = runner.invoke(s3_credential_create, [local_user])
        assert result.exit_code == 1
        assert "s3 connector" in result.output


def test_s3_credential_create_domain_not_allowed(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_get_auth_api: None,
) -> None:
    """
    When called with an identity trying to be mapped from a domain not in the
    allowed list, emit an error.
    """
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        bad_domain = "disallowed.org"
        assert migration_plan_s3.credentials[0].username is not None
        username = migration_plan_s3.credentials[0].username.split("@")[0]
        result = runner.invoke(s3_credential_create, ["--domain", bad_domain, username])
        assert result.exit_code == 1
        assert f"is not configured to allow domain {bad_domain}" in result.output


def test_s3_credential_create_no_domains(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_get_auth_api: None,
) -> None:
    """
    When called with a storage gateway without any domains configured, emit an
    error.
    """
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        assert migration_plan_s3.storage_gateway is not None
        migration_plan_s3.storage_gateway.allowed_domains = []
        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        assert migration_plan_s3.credentials[0].username is not None
        username = migration_plan_s3.credentials[0].username.split("@")[0]
        result = runner.invoke(s3_credential_create, [username])
        assert result.exit_code == 1
        assert "no domains configured" in result.output


def test_s3_credential_create_ambiguous_domains(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_get_auth_api: None,
) -> None:
    """
    When called with a storage gateway with multiple domains configured and none
    selected via command-line options, emit an error.
    """
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        assert migration_plan_s3.storage_gateway is not None

        migration_plan_s3.storage_gateway.allowed_domains = [
            "domain1.org",
            "domain2.org",
        ]
        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        assert migration_plan_s3.credentials[0].username is not None
        username = migration_plan_s3.credentials[0].username.split("@")[0]
        result = runner.invoke(s3_credential_create, [username])
        assert result.exit_code == 1
        assert "multiple domains" in result.output


def test_s3_credential_create_no_map_file_parser(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_get_auth_api: None,
) -> None:
    """
    When called with a migration plan without an s3 mapfile, emit an error.
    """
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        assert migration_plan_s3.storage_gateway is not None
        migration_plan_s3.storage_gateway.allowed_domains = [
            "domain1.org",
            "domain2.org",
        ]
        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        assert migration_plan_s3.credentials[0].username is not None
        username = migration_plan_s3.credentials[0].username.split("@")[0]
        result = runner.invoke(s3_credential_create, [username])
        assert result.exit_code == 1
        assert "multiple domains" in result.output


@pytest.mark.parametrize("pass_username", [False, True])
def test_s3_credential_create_user_has_s3_map_entry(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_identity_document: Callable[..., IdentityDocumentDict],
    mock_get_auth_api: None,
    mock_get_identities_response: Callable[[IdentityDocumentDict], None],
    pass_username: bool,
    mock_users: Callable[[List[str]], Dict[str, Any]],
    s3_process_user: str,
) -> None:
    """
    When called with an identity that has an s3 map entry but no credential
    in the migration plan, create a new credential
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump_credentials"):
        local_username = s3_map_entries[1].username
        assert migration_plan_s3.storage_gateway is not None
        assert migration_plan_s3.storage_gateway.allowed_domains is not None
        domain = migration_plan_s3.storage_gateway.allowed_domains[0]
        identity_documents = [
            mock_identity_document(me.username, domain=domain) for me in s3_map_entries
        ]
        for identity_document in identity_documents:
            mock_get_identities_response(identity_document)

        for cred in migration_plan_s3.credentials:
            assert cred.username != f"{local_username}@{domain}"

        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        local_user = s3_map_entries[1].username
        globus_username = identity_document["username"]

        if pass_username:
            result = runner.invoke(s3_credential_create, [local_user])
        else:
            mock_users([local_user, "other_user"])
            result = runner.invoke(s3_credential_create)

        found_cred = False
        for cred in migration_plan_s3.credentials:
            s3_cred = cast(S3Credential, cred)
            if (
                s3_cred.username == f"{local_username}@{domain}"
                and s3_cred.identity_username == globus_username
            ):
                assert s3_cred.s3_key_id == s3_map_entries[1].access_key_id
                assert s3_cred.s3_secret_key == s3_map_entries[1].secret_key
                found_cred = True
        assert found_cred
        assert result.exit_code == 0


def test_s3_credential_create_user_lacks_s3_map_entry(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_identity_document: Callable[[str], IdentityDocumentDict],
    mock_get_auth_api: None,
    mock_get_identities_response: Callable[[IdentityDocumentDict], None],
) -> None:
    """
    When called with an identity that does not have an s3 map entry
    in the migration plan, do not create a new credential
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump_credentials"):
        local_username = "unknown"
        identity_document = mock_identity_document(local_username)
        globus_username = identity_document["username"]
        mock_get_identities_response(identity_document)

        for cred in migration_plan_s3.credentials:
            assert cred.username != local_username

        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        result = runner.invoke(s3_credential_create, [local_username])
        found_cred = False
        for cred in migration_plan_s3.credentials:
            s3_cred = cast(S3Credential, cred)
            if (
                s3_cred.username == local_username
                and s3_cred.identity_username == globus_username
            ):
                assert s3_cred.s3_key_id == s3_map_entries[1].access_key_id
                assert s3_cred.s3_secret_key == s3_map_entries[1].secret_key
                found_cred = True
        assert not found_cred
        assert result.exit_code == 1
        assert "No usable S3 mapfile entry" in result.output


def test_s3_credential_create_user_lacks_globus_account(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_identity_document: Callable[..., IdentityDocumentDict],
    mock_get_auth_api: None,
    mock_get_identities_response: Callable[[IdentityDocumentDict], None],
) -> None:
    """
    When called with an identity that does not exist in auth do not create a
    new credential
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump_credentials"):
        local_username = s3_map_entries[1].username
        assert migration_plan_s3.storage_gateway is not None
        assert migration_plan_s3.storage_gateway.allowed_domains is not None
        domain = migration_plan_s3.storage_gateway.allowed_domains[0]
        identity_document = mock_identity_document(local_username, domain=domain)

        for cred in migration_plan_s3.credentials:
            assert cred.username != f"{local_username}@{domain}"

        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        local_user = s3_map_entries[1].username
        globus_username = identity_document["username"]
        result = runner.invoke(s3_credential_create, [local_user])
        found_cred = False
        for cred in migration_plan_s3.credentials:
            s3_cred = cast(S3Credential, cred)
            if (
                s3_cred.username == f"{local_username}@g{domain}"
                and s3_cred.identity_username == globus_username
            ):
                assert s3_cred.s3_key_id == s3_map_entries[1].access_key_id
                assert s3_cred.s3_secret_key == s3_map_entries[1].secret_key
                found_cred = True
        assert not found_cred
        assert result.exit_code == 1
        assert "Unable to resolve" in result.output


@pytest.mark.parametrize(
    "replace_key",
    [True, False],
)
def test_s3_credential_create_key_exists(
    gcs_config: GCSConfig,
    gridftp_config_parser_s3: GridFTPOptionParser,
    migration_plan_s3: MigrationPlan,
    s3_config_file: str,
    s3_map_entries: List[S3MapEntry],
    s3_map_files: List[str],
    mock_identity_document: Callable[..., IdentityDocumentDict],
    mock_get_auth_api: None,
    mock_get_identities_response: Callable[[IdentityDocumentDict], None],
    replace_key: bool,
) -> None:
    """
    When called with an identity that has an existing credential that has a
    different key, prompt and conditionally replace the key. If response is no,
    the key remains unchanged.
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump_credentials"):
        local_username = s3_map_entries[0].username

        assert migration_plan_s3.storage_gateway is not None
        assert migration_plan_s3.storage_gateway.allowed_domains is not None
        domain = migration_plan_s3.storage_gateway.allowed_domains[0]

        identity_document = mock_identity_document(local_username, domain=domain)
        mock_get_identities_response(identity_document)

        for cred in migration_plan_s3.credentials:
            s3_cred = cast(S3Credential, cred)
            assert s3_cred.username is not None
            s3_cred.username = s3_cred.username.replace("@globus.org", f"@{domain}")
            s3_cred.identity_username = cred.username
            s3_cred.s3_key_id = "old-key-id"
            s3_cred.s3_secret_key = "old-secret-key"

        mock_migration_load.return_value = migration_plan_s3
        runner = CliRunner()
        local_user = s3_map_entries[0].username
        globus_username = identity_document["username"]
        result = runner.invoke(
            s3_credential_create,
            [local_user],
            input="y\n" if replace_key else "n\n",
        )
        found_cred = False
        for cred in migration_plan_s3.credentials:
            s3_cred = cast(S3Credential, cred)
            if (
                s3_cred.username == f"{local_username}@{domain}"
                and s3_cred.identity_username == globus_username
            ):
                assert (
                    s3_cred.s3_key_id == s3_map_entries[0].access_key_id
                ) == replace_key
                assert (
                    s3_cred.s3_secret_key == s3_map_entries[0].secret_key
                ) == replace_key
                found_cred = True
        assert found_cred
        assert result.exit_code == 0
