import os
import pathlib
from typing import Any, Callable, Dict, List, Tuple

import pytest

from gcs_migration.parsers import GridFTPOptionParser

from tests.fixtures.gridftp_config import (
    GridFTPConfigFixture,
    GridFTPConfigFixtureGenerator,
    GridFTPConfigPolicy,
)


GridFTPConfFixtureType = Tuple[str, Dict]

# Override fixtures that use FakeFilesystem with tmp_path so that the gridftp
# server lib can find the actual files
@pytest.fixture
def gridftp_conf(tmp_path: pathlib.Path) -> GridFTPConfigPolicy:
    name = tmp_path / "gridftp.conf"

    class GridftpConf:
        options = {"port": 2811}
        filename = str(name)

        def __str__(self) -> str:
            return f"""\
# globus-gridftp-server configuration file

# this is a comment

# option names beginning with '$' will be set as environment variables, e.g.
# $GLOBUS_ERROR_VERBOSE 1
# $GLOBUS_TCP_PORT_RANGE 50000,51000

# port
port {self.options["port"]}
"""

    conf = GridftpConf()

    with open(name, "w") as f:
        print(str(conf), file=f)

    return conf


@pytest.fixture
def gridftp_config_d(tmp_path: pathlib.Path) -> GridFTPConfigFixtureGenerator:
    def _gridftp_config_d(configs: List[GridFTPConfigPolicy]) -> GridFTPConfigFixture:
        gridftp_d = tmp_path / "gridftp.d"
        if not os.path.exists(gridftp_d):
            os.makedirs(gridftp_d, 0o0755)

        options: Dict[str, Any] = {}
        for config in configs:
            with open(gridftp_d / config.filename, "w") as f:
                f.write(str(config))
            options = {**options, **config.options}

        return (gridftp_d, options)

    return _gridftp_config_d


def missing_gridftp_library() -> bool:
    try:
        import ctypes
        import ctypes.util

        gridftp_server = ctypes.CDLL(ctypes.util.find_library("globus_gridftp_server"))
        if gridftp_server and hasattr(gridftp_server, "globus_i_gfs_module"):
            return False
    except Exception:
        pass
    return True


@pytest.mark.skipif(missing_gridftp_library(), reason="Missing GridFTP Library")
def test_parse_gridftp_config(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
) -> None:
    configs = [
        gridftp_config_gcs,
        gridftp_config_credential,
        gridftp_config_logging,
    ]

    config_dir, config_dir_options = gridftp_config_d(configs)

    config_data = {**gridftp_conf.options, **config_dir_options}

    options = GridFTPOptionParser(
        config_file=gridftp_conf.filename, config_dir=config_dir
    )

    for config, value in config_data.items():
        parsed_value = options.get_option(config)
        assert parsed_value == value

    assert options.get_option("detach") is False
    assert options.get_option("daemon") is True
    assert options.get_option("chdir_to") is None

    with pytest.raises(Exception):
        options.get_option("bogus")


def test_parse_gridftp_config_int(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> None:
    class IntGridFTPConfigPolicy:
        options = {
            "threads": 2,
        }
        filename = "threads"

        def __str__(self) -> str:
            return "threads 2"

    threads_config = IntGridFTPConfigPolicy()
    config_dir, config_dir_options = gridftp_config_d([threads_config])
    gridftp_config = config_dir, config_dir_options
    options = fake_gridftp_config_parser_creator(gridftp_config)

    assert options.get_option("threads") == 2
    assert options.get_int_option("threads") == 2
    with pytest.raises(TypeError):
        options.get_string_option("threads")
    with pytest.raises(TypeError):
        options.get_bool_option("threads")


def test_parse_gridftp_config_bool(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> None:
    class BoolGridFTPConfigPolicy:
        options = {
            "allow_udt": True,
        }
        filename = "udt"

        def __str__(self) -> str:
            return "allow_udt"

    udt_config = BoolGridFTPConfigPolicy()
    config_dir, config_dir_options = gridftp_config_d([udt_config])
    gridftp_config = config_dir, config_dir_options
    options = fake_gridftp_config_parser_creator(gridftp_config)

    assert options.get_option("allow_udt") is True
    assert options.get_bool_option("allow_udt") is True
    with pytest.raises(TypeError):
        assert options.get_int_option("allow_udt")
    with pytest.raises(TypeError):
        assert options.get_string_option("allow_udt")


def test_parse_gridftp_config_string(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> None:
    class StringGridFTPConfigPolicy:
        options = {
            "load_dsi_module": "s3",
        }
        filename = "s3"

        def __str__(self) -> str:
            return "load_dsi_module s3"

    s3_config = StringGridFTPConfigPolicy()
    config_dir, config_dir_options = gridftp_config_d([s3_config])
    gridftp_config = config_dir, config_dir_options
    options = fake_gridftp_config_parser_creator(gridftp_config)

    assert options.get_option("load_dsi_module") == "s3"
    assert options.get_string_option("load_dsi_module") == "s3"
    with pytest.raises(TypeError):
        assert options.get_int_option("load_dsi_module")
    with pytest.raises(TypeError):
        assert options.get_bool_option("load_dsi_module")


def test_parse_gridftp_config_undefined_option(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> None:
    config_dir, config_dir_options = gridftp_config_d([])
    gridftp_config = config_dir, config_dir_options
    options = fake_gridftp_config_parser_creator(gridftp_config)

    with pytest.raises(KeyError):
        assert options.get_option("non_existent") == "s3"
    with pytest.raises(KeyError):
        assert options.get_string_option("non_existent") == "s3"
    with pytest.raises(KeyError):
        assert options.get_bool_option("non_existent") == "s3"
    with pytest.raises(KeyError):
        assert options.get_int_option("non_existent") == "s3"
